"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Time = void 0;
/** @fileOverview a time class **/
var timeutil_1 = require("./timeutil");
var NEGATIVE_SUPPORT = false; // A later version will suport negatives
var Time = /** @class */ (function () {
    function Time(timeExpression) {
        if (timeExpression === void 0) { timeExpression = "0"; }
        this.__sign = 1;
        this._time = [];
        this.set(timeExpression);
    }
    Object.defineProperty(Time.prototype, "_sign", {
        /**
         * @ignore
         */
        get: function () {
            return this.__sign;
        },
        set: function (val) {
            // TODO: Currently, negatives are not supported. Negatives use a _sign parameter set to -1
            // For now, this setter is used to detect a negative, and throws accordingly. When negatives
            // are supported, this setter can be removed
            if (val === -1) {
                throw new RangeError('negatives are not supported');
            }
            this.__sign = val;
        },
        enumerable: false,
        configurable: true
    });
    Time.prototype.set = function (timeExpression) {
        /**
         * Immediately changes the instances time to the given argument.
         */
        if (typeof timeExpression === 'number') {
            if (timeExpression < 0) {
                this._sign = -1;
                timeExpression = Math.abs(timeExpression);
            }
            // @ts-ignore
            timeExpression = timeExpression.toString() + 'ms';
        }
        if (typeof timeExpression === 'string') {
            if (/^[^\d\w]*-/.test(timeExpression)) {
                this._sign = -1;
                timeExpression = timeExpression.replace(/^([^\d\w]*)(-)(.*)$/, '$1$3');
            }
            // match regex ending with dot
            var matchArray = timeExpression.match(/\.(.*)$/);
            if (matchArray) {
                var mills = matchArray[1];
                if (mills.length === 2) {
                    timeExpression += '0';
                }
                if (mills.length === 1) {
                    timeExpression += '00';
                }
            }
            this._time = this._parse(timeExpression).map(function (val) { return val === null ? [0] : [val]; });
        }
        else if (typeof timeExpression === 'number') {
        }
        else if (timeExpression instanceof Time) {
            this._sign = timeExpression._sign;
            this._time = timeExpression._time.map(function (val) { return [val[0]]; });
        }
        else if (typeof timeExpression === 'object') {
            this._time = [
                [((timeExpression === null || timeExpression === void 0 ? void 0 : timeExpression.years) || 0)],
                [((timeExpression === null || timeExpression === void 0 ? void 0 : timeExpression.days) || 0)],
                [((timeExpression === null || timeExpression === void 0 ? void 0 : timeExpression.hours) || 0)],
                [((timeExpression === null || timeExpression === void 0 ? void 0 : timeExpression.minutes) || 0)],
                [((timeExpression === null || timeExpression === void 0 ? void 0 : timeExpression.seconds) || 0)],
                [((timeExpression === null || timeExpression === void 0 ? void 0 : timeExpression.milliseconds) || 0)],
            ];
        }
        else {
            throw new TypeError("Can't parse given timeExpression.");
        }
        this._adjustTime(0);
    };
    Time.prototype.add = function (time, _ignoreSigns) {
        if (_ignoreSigns === void 0) { _ignoreSigns = false; }
        var toAdd = time instanceof Time ? time : new Time(time);
        if (!_ignoreSigns) {
            if (this._sign === -1 && toAdd._sign === 1) {
                var signTemp_1 = this._sign;
                this.abs(true).subtract(toAdd);
                this._sign *= signTemp_1;
                return;
            }
            if (this._sign === 1 && toAdd._sign === -1) {
                var signTemp_2 = this._sign;
                this.subtract(toAdd.abs());
                this._sign *= signTemp_2;
            }
        }
        var signTemp = this._sign;
        this.abs(true)._adjustTime(toAdd.abs().inMilliseconds());
        this._sign = signTemp;
        return;
        this.milliseconds = this.milliseconds - toAdd.milliseconds;
        this.seconds = this.seconds - toAdd.seconds;
        this.minutes = this.minutes - toAdd.minutes;
        this.hours = this.hours - toAdd.hours;
        this.days = this.days - toAdd.days;
        this.years = this.years - toAdd.years;
    };
    Time.prototype.abs = function (set) {
        if (set === void 0) { set = false; }
        /** Returns a new Time instance that is the absolute value.
         *  If 'set' is true, mokes this instance absolute value and
         *  returns that instead.
         *  @ignore - not needed until negatives are supported
         */
        if (set) {
            this._sign = 1;
            return this;
        }
        var absTime = new Time(this);
        absTime._sign = 1;
        return absTime;
    };
    Time.prototype.setSign = function (sign) {
        /**
         * Reserved for future features with negatives
         * @ignore
         */
        // @ts-ignore
        this._sign = Math.sign(sign) || 1;
        return this;
    };
    Time.prototype.equals = function (time) {
        var compare = time instanceof Time ? time : new Time(time);
        return this.years === compare.years
            && this.days === compare.days
            && this.hours === compare.hours
            && this.minutes === compare.minutes
            && this.seconds === compare.seconds
            && this.milliseconds === compare.milliseconds;
    };
    Time.prototype.gt = function (time) {
        var compare = time instanceof Time ? time : new Time(time);
        if (this.years === compare.years) {
            if (this.days === compare.days) {
                if (this.hours === compare.hours) {
                    if (this.minutes === compare.minutes) {
                        if (this.seconds === compare.seconds) {
                            if (this.milliseconds === compare.milliseconds) {
                                return false;
                            }
                            return this.milliseconds > compare.milliseconds;
                        }
                        return this.seconds > compare.seconds;
                    }
                    return this.minutes > compare.minutes;
                }
                return this.hours > compare.hours;
            }
            return this.days > compare.days;
        }
        return this.years > compare.years;
    };
    Time.prototype.gte = function (time) {
        var compare = time instanceof Time ? time : new Time(time);
        if (this.years === compare.years) {
            if (this.days === compare.days) {
                if (this.hours === compare.hours) {
                    if (this.minutes === compare.minutes) {
                        if (this.seconds === compare.seconds) {
                            if (this.milliseconds === compare.milliseconds) {
                                return true;
                            }
                            return this.milliseconds > compare.milliseconds;
                        }
                        return this.seconds > compare.seconds;
                    }
                    return this.minutes > compare.minutes;
                }
                return this.hours > compare.hours;
            }
            return this.days > compare.days;
        }
        return this.years > compare.years;
    };
    Time.prototype.lt = function (time) {
        var compare = time instanceof Time ? time : new Time(time);
        if (this.years === compare.years) {
            if (this.days === compare.days) {
                if (this.hours === compare.hours) {
                    if (this.minutes === compare.minutes) {
                        if (this.seconds === compare.seconds) {
                            if (this.milliseconds === compare.milliseconds) {
                                return false;
                            }
                            return this.milliseconds < compare.milliseconds;
                        }
                        return this.seconds < compare.seconds;
                    }
                    return this.minutes < compare.minutes;
                }
                return this.hours < compare.hours;
            }
            return this.days < compare.days;
        }
        return this.years < compare.years;
    };
    Time.prototype.lte = function (time) {
        var compare = time instanceof Time ? time : new Time(time);
        if (this.years === compare.years) {
            if (this.days === compare.days) {
                if (this.hours === compare.hours) {
                    if (this.minutes === compare.minutes) {
                        if (this.seconds === compare.seconds) {
                            if (this.milliseconds === compare.milliseconds) {
                                return true;
                            }
                            return this.milliseconds < compare.milliseconds;
                        }
                        return this.seconds < compare.seconds;
                    }
                    return this.minutes < compare.minutes;
                }
                return this.hours < compare.hours;
            }
            return this.days < compare.days;
        }
        return this.years < compare.years;
    };
    Time.prototype.subtract = function (time, _ignoreSigns) {
        if (_ignoreSigns === void 0) { _ignoreSigns = false; }
        var toSubtract = time instanceof Time ? time : new Time(time);
        if (!_ignoreSigns) {
            if (this._sign === 1 && toSubtract._sign === -1) {
                this.add(toSubtract.abs());
                return;
            }
            if (this._sign === -1 && toSubtract._sign === 1) {
                var signTemp = this._sign;
                this.abs(true).add(toSubtract);
                this._sign *= signTemp;
                return;
            }
            if (this._sign + toSubtract._sign === -2) {
                var signTemp = this._sign;
                this.abs(true).subtract(toSubtract, true);
                this._sign *= signTemp;
                return;
            }
        }
        this._adjustTime(-toSubtract.inMilliseconds());
        return;
        this.milliseconds = this.milliseconds - toSubtract.milliseconds;
        this.seconds = this.seconds - toSubtract.seconds;
        this.minutes = this.minutes - toSubtract.minutes;
        this.hours = this.hours - toSubtract.hours;
        this.days = this.days - toSubtract.days;
        this.years = this.years - toSubtract.years;
        return;
        if (toSubtract.abs().gt(this.abs())) {
            // TODO: replace with setTime
            var newSubtract = new Time(this);
            this.years = toSubtract.years;
            this.hours = toSubtract.hours;
            this.days = toSubtract.days;
            this.minutes = toSubtract.minutes;
            this.seconds = toSubtract.seconds;
            this.milliseconds = toSubtract.milliseconds;
            this._sign *= -1;
            return this.subtract(newSubtract);
        }
        this.years = Math.abs(this.years) - Math.abs(toSubtract.years);
        this.days = Math.abs(this.days) - Math.abs(toSubtract.days);
        this.hours = Math.abs(this.hours) - Math.abs(toSubtract.hours);
        this.minutes = Math.abs(this.minutes) - Math.abs(toSubtract.minutes);
        this.seconds = Math.abs(this.seconds) - Math.abs(toSubtract.seconds);
        this.milliseconds = Math.abs(this.milliseconds) - Math.abs(toSubtract.milliseconds);
    };
    Time.prototype.inMilliseconds = function () {
        return this.years * 365 * 24 * 60 * 60 * 1000 +
            this.days * 24 * 60 * 60 * 1000 +
            this.hours * 60 * 60 * 1000 +
            this.minutes * 60 * 1000 +
            this.seconds * 1000 +
            this.milliseconds;
    };
    Time.prototype._adjustTime = function (milliseconds) {
        /** Adjusts time by a number of milliseconds. Pass negative number to decrement.
         */
        var aac = this._adjustAndCarry;
        aac(this._time[0], Infinity, aac(this._time[1], 364, aac(this._time[2], 23, aac(this._time[3], 59, aac(this._time[4], 59, aac(this._time[5], 999, milliseconds))))));
    };
    Time.prototype._adjustAndCarry = function (num, resetValue, interval) {
        var val = num[0] + interval;
        var carry = 0;
        while (val < 0) {
            val += (resetValue + 1);
            carry -= 1;
        }
        while (val > resetValue) {
            val -= (resetValue + 1);
            carry += 1;
        }
        num[0] = val;
        return carry;
    };
    Time.prototype._getCarryover = function (num, threshold) {
        /** Calculates the "carry-over", that is, how many whole units can be divided from a
         * threshold. This is used to calculate when, say, the seconds unit has more than allowed (60).
         * In that case, a number of "minutes" should be extracted from it (over the "threshold" of 60).
         */
        var carryover = (num - (num % threshold)) / threshold;
        var remaining = num % threshold;
        return [carryover, remaining];
    };
    Time.prototype._fromMilliseconds = function (num) {
        var _a, _b, _c, _d, _e;
        var result = [0, 0, 0, 0, 0, num];
        if (result[5] < 1000) {
            return result;
        }
        _a = this._getCarryover(result[5], 1000), result[4] = _a[0], result[5] = _a[1];
        if (result[4] < 60) {
            return result;
        }
        _b = this._getCarryover(result[4], 60), result[3] = _b[0], result[4] = _b[1];
        var minutes = result[3];
        if (result[3] < 60) {
            return result;
        }
        _c = this._getCarryover(result[3], 60), result[2] = _c[0], result[3] = _c[1];
        var hours = result[2];
        if (hours < 24) {
            return result;
        }
        _d = this._getCarryover(result[2], 24), result[1] = _d[0], result[2] = _d[1];
        var days = result[1];
        if (result[1] < 365) {
            return result;
        }
        _e = this._getCarryover(result[1], 365), result[0] = _e[0], result[1] = _e[1];
        return result;
    };
    Time.prototype._adjustOverflow = function (place, threshold) {
        /**
         * Sets the value at placemarker (value) to a positive number within its
         * threshold, adding to larger places if needed.
         *
         * If the value is within the threshold, nothing happens.
         */
        if (place === 0) {
            // the largest placement has no limit
            return 0;
        }
        if (this._time[place][0] >= threshold) {
            var _a = this._getCarryover(this._time[place][0], threshold), carryOver = _a[0], remaining = _a[1];
            this._time[place][0] = remaining;
            return carryOver;
        }
        return 0;
    };
    Time.prototype._adjustUnderflow = function (place, threshold) {
        if (this._time[place][0] < 0) {
            if (this._time.slice(0, place).some(function (el) { return el[0]; })) {
                var _a = this._getCarryover(Math.abs(this._time[place][0]), threshold), carryOver = _a[0], remainingInverse = _a[1];
                this._time[place][0] = (threshold - remainingInverse) % threshold;
                return -(carryOver);
            }
            // Arriving here, there are no further places to borrow from, and sign must flip
            if (NEGATIVE_SUPPORT) {
                this._sign *= -1;
                this._time[place][0] = Math.abs(this._time[place][0]);
            }
            else {
                this._time[place][0] = 0;
            }
        }
        return 0;
    };
    Time.prototype._largestIndex = function () {
        /**
         * Returns the index (from this._time) which is the largest unit of
         * time that is non-zero.
         *
         * It is important to know the largest unit, because if it is subtracted
         * by a value greater, the sign needs to be flipped.
         */
        for (var i = 0; i < this._time.length; i++) {
            if (this._time[i][0]) {
                return i;
            }
        }
        return this._time.length - 1;
    };
    Object.defineProperty(Time.prototype, "milliseconds", {
        // [_years, _days, _hours, minutes, seconds, milliseconds]
        //  0      1     2      3        4        5
        get: function () {
            return this._time[5][0] * this._sign;
        },
        set: function (n) {
            this._time[5][0] = n;
            var over = this._adjustOverflow(5, 1000);
            var under = this._adjustUnderflow(5, 1000);
            this.seconds = this._time[4][0] + over + under;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(Time.prototype, "ms", {
        /**
         * @deprecated use [[Time.milliseconds]] instead
         */
        get: function () {
            /** Alias of .milliseconds. **/
            return this.milliseconds;
        },
        set: function (n) {
            this.milliseconds = n;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(Time.prototype, "seconds", {
        get: function () {
            return this._time[4][0] * this._sign;
        },
        set: function (n) {
            this._time[4][0] = n;
            var over = this._adjustOverflow(4, 60);
            var under = this._adjustUnderflow(4, 60);
            this.minutes = this._time[3][0] + over + under;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(Time.prototype, "secs", {
        get: function () {
            /** Alias of this#seconds **/
            return this.seconds;
        },
        set: function (n) {
            /**
             * @alias seconds
             */
            this.seconds = n;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(Time.prototype, "minutes", {
        get: function () {
            return this._time[3][0] * this._sign;
        },
        set: function (n) {
            this._time[3][0] = n;
            var over = this._adjustOverflow(3, 60);
            var under = this._adjustUnderflow(3, 60);
            this.hours = this._time[2][0] + over + under;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(Time.prototype, "mins", {
        /**
         * @deprecated use [[Time.minutes]] instead
         **/
        get: function () {
            return this.minutes;
        },
        set: function (n) {
            this.minutes = n;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(Time.prototype, "hours", {
        get: function () {
            return this._time[2][0] * this._sign;
        },
        set: function (n) {
            this._time[2][0] = n;
            var over = this._adjustOverflow(2, 24);
            var under = this._adjustUnderflow(2, 24);
            this.days = this._time[1][0] + over + under;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(Time.prototype, "days", {
        get: function () {
            return this._time[1][0] * this._sign;
        },
        set: function (n) {
            this._time[1][0] = n;
            var over = this._adjustOverflow(1, 365);
            var under = this._adjustUnderflow(1, 365);
            this.years = this._time[0][0] + over + under;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(Time.prototype, "years", {
        get: function () {
            return this._time[0][0] * this._sign;
        },
        set: function (n) {
            if (n < 0) {
                this._sign *= -1;
                n = Math.abs(n);
            }
            this._time[0][0] = n;
        },
        enumerable: false,
        configurable: true
    });
    Time.prototype._parse = function (time) {
        // TODO: this does NOT parse negative values. That is expected to be handled
        // by the instances _sign prop. A seperate parseTime for the client should
        // be created, which applies the _sign as appropriate after the initial parsing.
        if (/[dshmy]+/.test(time)) {
            var valuesAtIndex = ['y', 'd', 'h', 'm', 's', 'ms'];
            var parsedTime = [null, null, null, null, null, null];
            var workingInt = '';
            for (var i = 0; i < time.length; i++) {
                if (/\d/.test(time[i])) {
                    workingInt += time[i];
                }
                else if (time[i] + time[i + 1] === 'ms') {
                    if (parsedTime[5] !== null) {
                        throw new SyntaxError("Duplicate token 'ms'");
                    }
                    parsedTime[5] = workingInt;
                    workingInt = '';
                    i++;
                }
                else if (/[dshmy]/.test(time[i])) {
                    if (parsedTime[valuesAtIndex.indexOf(time[i])] !== null) {
                        throw new SyntaxError("Duplicate token " + time[i]);
                    }
                    parsedTime[valuesAtIndex.indexOf(time[i])] = workingInt;
                    workingInt = '';
                }
                else {
                    throw new SyntaxError("Unexpected token " + time[i]);
                }
            }
            return parsedTime.map(function (i) { return isNaN(parseInt(i)) ? 0 : parseInt(i); });
        }
        if (/(?:\d+:){0,4}\d+/.test(time)) {
            // colon separated syntax
            var _a = time.split('.'), others = _a[0], mils = _a[1];
            var parsed = others.split(/[^\d.]/).concat(mils);
            parsed = parsed.map(function (i) { return isNaN(parseInt(i)) ? 0 : parseInt(i); });
            while (parsed.length < 6) {
                parsed = [0].concat(parsed);
            }
            // TODO milliseconds
            return parsed;
        }
        else {
            throw TypeError("Cannot _parse string as time.");
        }
    };
    /** Returns time in "YY:DD:hh:ss:mm.mills" format, optionally truncating some units.
     *
     *  @param minUnit - Lowest unit of time to show. Units lower than `minUnit` will not be shown. 'ms' or '' means
     *  include every unit.
     *  @param maxUnit Largest unit of time to show. All units of time larger than this are *added* to the maxUnit.
     *  E.g. '01:01:01.000'.toString('', 'm') returns '61:01.000'
     *
     *  @returns Formatted string representing the time. If maxUnit is greater than minUnit, returns ''
     */
    Time.prototype.toString = function (minUnit, maxUnit) {
        if (maxUnit === void 0) { maxUnit = 'y'; }
        minUnit = minUnit || 'ms';
        var result = [];
        var timeUnitIndex = ['y', 'd', 'h', 'm', 's', 'ms'];
        if (timeUnitIndex.indexOf(minUnit) < timeUnitIndex.indexOf(maxUnit)) {
            return '';
        }
        //HACK: If maxUnit is ms, just return from .inMilliseconds()
        if (maxUnit === 'ms') {
            return this.inMilliseconds().toString();
        }
        var unitDownMultiplier = [365, 24, 60, 60, 1000, 1];
        var nonZeroSeen = false;
        var firstUnit = timeUnitIndex.indexOf(maxUnit);
        var carryOver = 0;
        for (var i = 0; i < this._time.length - 1; i++) {
            var val = this._time[i][0] + carryOver;
            carryOver = 0;
            // seconds (i=4) should always be added
            if (val || nonZeroSeen || i === 4) {
                nonZeroSeen = true;
                if (i < firstUnit) {
                    carryOver += val * unitDownMultiplier[i];
                }
                else {
                    result.push(timeutil_1.addPadding(val, 2));
                }
            }
        }
        var lastUnit = timeUnitIndex.indexOf(minUnit) - firstUnit;
        var resultStr = result.slice(0, lastUnit + 1).join(':');
        if (!minUnit || minUnit === 'ms') {
            resultStr += "." + timeutil_1.addPadding(this.milliseconds, 3);
        }
        return resultStr;
    };
    return Time;
}());
exports.Time = Time;
//# sourceMappingURL=time.js.map