"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (Object.prototype.hasOwnProperty.call(b, p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Timer = void 0;
var time_1 = require("./time");
var timeutil_1 = require("./timeutil");
/**
 * The main timer class.
 *
 * Example:
 * ```javascript
 * new Timer('3:00', () => console.log('Time\'s up'));
 * ```
 *
 * See {@link Timer.constructor} for more uses
 */
var Timer = /** @class */ (function (_super) {
    __extends(Timer, _super);
    function Timer(timeExpression) {
        var args = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            args[_i - 1] = arguments[_i];
        }
        var _this = _super.call(this, timeExpression) || this;
        _this.options = {
            divider: ":",
            repeat: 0,
            immediateInterval: false,
            finalInterval: true,
            interval: 1,
            setInterval: timeutil_1.setGoodInterval,
            clearInterval: timeutil_1.clearGoodInterval
        };
        /**
         * whether or not the timer is paused.
         */
        _this.isPaused = false;
        _this._startMarker = -1;
        //// Functions for backwards compatibility with t-minus 1.0 ////
        _this.resume = _this.unpause;
        _this.play = _this.unpause;
        _this.time = new time_1.Time(timeExpression);
        _this.initialTime = _this.time.toString();
        _this.adjustTime(0);
        // timeExpression, onTimeout, onInterval sig
        if (typeof args[0] === 'function' && typeof args[1] === 'function') {
            _this.options = __assign(__assign(__assign({}, _this.options), { onTimeout: args[0], onInterval: args[1] }), (args[2] || {}));
        }
        else if (typeof args[0] === 'function') {
            // @ts-ignore
            _this.options = __assign(__assign(__assign({}, _this.options), { onTimeout: args[0] }), (args[1] || {}));
        }
        else if (typeof args[0] === 'object') {
            _this.options = __assign(__assign({}, _this.options), args[0]);
        }
        // @ts-ignore
        if (_this.options.repeat === true) {
            _this.options.repeat = Infinity;
        }
        // @ts-ignore
        if (_this.options.repeat === false) {
            _this.options.repeat = 0;
        }
        _this.isPaused = !!_this.options.startPaused;
        _this._startIntervalLoop(_this.options.immediateInterval);
        return _this;
    }
    Object.defineProperty(Timer.prototype, "initialTime", {
        get: function () {
            return this._initialTime;
        },
        set: function (val) {
            if (typeof val !== 'string') {
                throw TypeError('initialTime must be a string');
            }
            this._initialTime = val;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(Timer.prototype, "milliseconds", {
        get: function () {
            if (!this.isPaused) {
                return (Date.now() - this.lastTick) % 1000;
            }
            return this.time["milliseconds"] % 1000;
        },
        set: function (val) {
            this.time.milliseconds = val;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(Timer.prototype, "intervalId", {
        get: function () {
            return this._intervalId;
        },
        set: function (val) {
            if (this._intervalId) {
                // @ts-ignore
                this.options.clearInterval(this._intervalId);
            }
            this._intervalId = val;
        },
        enumerable: false,
        configurable: true
    });
    /**
     * Manually decrement the timer by its interval (by default 1 second), triggering the [[Timer.onInterval]] if defined, unless paused.
     * @param force do this even if paused (timer will still remain paused)
     */
    Timer.prototype.tick = function (force) {
        if (force === void 0) { force = false; }
        /** The main loop on the timer. **/
        // if an interval triggers a tick that was supposed to stop (due to quickly pausing again), it's ignored completely;
        // the same tick will rerun when the timer correctly resumes
        if (this.isPaused && !force) {
            return;
        }
        this.lastTick = Date.now();
        // Edge Case: timer with interval > 1 may pass "timesUp" (all 0's) rather than land on it.
        // When only seconds remain and are less than interval, adjust to be the interval exactly to get to 0.
        if (!this.years && !this.days && !this.hours && !this.minutes && this.seconds < this.options.interval) {
            this.seconds = this.options.interval;
        }
        this.adjustTime(-this.options.interval);
        if (!this.years && !this.days && !this.hours && !this.minutes && !this.seconds) {
            //TODO: Will there ever be millisecond remaining? Should a timeout be set here in that case?
            // or will _mills always be 0 (and the case is handled on a resume)
            try {
                this.options.onInterval && this.options.finalInterval && this.options.onInterval.call(this, this);
                this.pause();
                this.options.onTimeout && this.options.onTimeout.call(this, this);
            }
            catch (e) {
                throw e;
            }
            finally {
                if (this.options.repeat) {
                    this.options.repeat--;
                    this.setFromString(this.initialTime);
                    this.unpause();
                }
            }
        }
        else {
            this.options.onInterval && this.options.onInterval.call(this, this);
        }
    };
    /**
     * Flips the timer to paused if unpaused, and to unpaused if paused.
     */
    Timer.prototype.togglePause = function () {
        if (this.isPaused) {
            return this.unpause();
        }
        else {
            return this.pause();
        }
    };
    /**
     * Pauses the timer. If the timer is already paused, this has no effect.
     */
    Timer.prototype.pause = function () {
        if (this.isPaused) {
            return;
        }
        var timePaused = Date.now();
        this.isPaused = true;
        this.remainingSec = timePaused - this.lastTick;
        this.milliseconds = this.remainingSec;
        // @ts-ignore
        clearInterval(this.intervalId);
        return this.isPaused;
    };
    /**
     * Unpauses the timer. If the timer is already unpaused, this has no effect.
     */
    Timer.prototype.unpause = function () {
        if (!this.isPaused) {
            return;
        }
        this.isPaused = false;
        this.lastTick = Date.now(); // brings us back to current time since we last paused.
        setTimeout(function () { this._startIntervalLoop(true); }.bind(this), this.remainingSec);
        return this.isPaused;
    };
    Timer.prototype._startIntervalLoop = function (initialTick) {
        // the timer *can* be paused before 1 second passes, so it must be checked.
        if (this.isPaused) {
            return;
        }
        initialTick && this.tick();
        this.lastTick = Date.now();
        this.intervalId = this.options.setInterval(this.tick.bind(this), this.options.interval * 1000);
    };
    Timer.prototype._parse = function (time) {
        // TODO: this does NOT parse negative values. That is expected to be handled
        // by the instances _sign prop. A seperate parseTime for the client should
        // be created, which applies the _sign as appropriate after the initial parsing.
        if (/[dshmy]+/.test(time)) {
            var valuesAtIndex = ['y', 'd', 'h', 'm', 's', 'ms'];
            var parsedTime = [null, null, null, null, null, null];
            var workingInt = '';
            for (var i = 0; i < time.length; i++) {
                if (/\d/.test(time[i])) {
                    workingInt += time[i];
                }
                else if (time[i] + time[i + 1] === 'ms') {
                    if (parsedTime[5] !== null) {
                        throw new SyntaxError("Duplicate token 'ms'");
                    }
                    parsedTime[5] = workingInt;
                    workingInt = '';
                    i++;
                }
                else if (/[dshmy]/.test(time[i])) {
                    if (parsedTime[valuesAtIndex.indexOf(time[i])] !== null) {
                        throw new SyntaxError("Duplicate token " + time[i]);
                    }
                    parsedTime[valuesAtIndex.indexOf(time[i])] = workingInt;
                    workingInt = '';
                }
                else {
                    throw new SyntaxError("Unexpected token " + time[i]);
                }
            }
            return parsedTime.map(function (i) { return isNaN(parseInt(i)) ? 0 : parseInt(i); });
        }
        if (/(?:\d+:){0,4}\d+/.test(time)) {
            // colon separated syntax
            var _a = time.split('.'), others = _a[0], mils = _a[1];
            var parsed = others.split(/[^\d.]/).concat(mils);
            parsed = parsed.map(function (i) { return isNaN(parseInt(i)) ? 0 : parseInt(i); });
            while (parsed.length < 6) {
                parsed = [0].concat(parsed);
            }
            // TODO milliseconds
            return parsed;
        }
        else {
            throw TypeError("Cannot _parse string as time.");
        }
    };
    /**
     * @deprecated use [[Timer.set]] instead
     *
     * Change the timer to a given string.
     * @param time an expression of a time.
     */
    Timer.prototype.setFromString = function (time) {
        this.set(time);
        // this.time = new Time(time);
    };
    Timer.prototype.adjustTime = function (seconds) {
        if (seconds === void 0) { seconds = -1; }
        /** Adjusts time by a number of seconds. Pass negative number to decrement.
         */
        this._adjustTime(seconds * 1000);
    };
    Timer.prototype.adjustAndCarry = function (num, resetValue, interval) {
        var val = num[0] + interval;
        var carry = 0;
        while (val < 0) {
            val += (resetValue + 1);
            carry -= 1;
        }
        while (val > resetValue) {
            val -= (resetValue + 1);
            carry += 1;
        }
        num[0] = val;
        return carry;
    };
    /**
     * Resets the timer back to its initial time. [[Timer.onTimeout]] does not get called.
     *
     * Does not change whether the timer is paused or not.
     */
    Timer.prototype.reset = function () {
        this.setFromString(this.initialTime);
    };
    /**
     * Like [[Timer.reset]], but unpauses the timer as well.
     */
    Timer.prototype.restart = function () {
        this.reset();
        this.unpause();
    };
    //// UI-Functions ////
    Timer.prototype._addPadding = function (number, zeros) {
        return timeutil_1.addPadding(number, zeros);
    };
    Timer.prototype._timeAsArray = function () {
        return [this.years, this.days, this.hours, this.minutes, this.seconds, this.milliseconds];
    };
    Timer.prototype.getMillisecondsUI = function (padding) {
        var dateOffset = this.lastTick ? Date.now() - this.lastTick : 0;
        return this._addPadding(dateOffset + this.milliseconds, padding);
    };
    Timer.prototype.getSecondsUI = function (padding) {
        if (padding === void 0) { padding = 2; }
        return this._addPadding(this.seconds, padding);
    };
    Timer.prototype.getMinutesUI = function (padding) {
        return this._addPadding(this.minutes, padding);
    };
    Timer.prototype.getHoursUI = function (padding) {
        return this._addPadding(this.hours, padding);
    };
    Timer.prototype.getDaysUI = function (padding) {
        return this._addPadding(this.days, padding);
    };
    Timer.prototype.getYearsUI = function (padding) {
        return this._addPadding(this.years, padding);
    };
    Timer.prototype.getFullTimeUI = function (includeMilliseconds) {
        if (includeMilliseconds === void 0) { includeMilliseconds = false; }
        var started = false;
        var result = [];
        var pad = this._addPadding;
        if (this.years) {
            started = true;
            result.push(this.years);
        }
        if (this.days || started) {
            started = true;
            result.push(this.days);
        }
        if (this.hours || started) {
            started = true;
            result.push(pad(this.hours, 2));
        }
        if (this.minutes || started) {
            started = true;
            result.push(pad(this.minutes, 2));
        }
        result.push(pad(this.seconds, 2));
        if (includeMilliseconds) {
            result.push(pad(this.milliseconds, 2));
        }
        return result.join(this.options.divider);
    };
    /**
     * Formats the current time as a string, depending on the value of fmt.
     *
     * @param fmt a template string that will be used to populate with time values.
     *
     * fmtTime looks for special characters in the string that start with `%`, depending on the next value, fmtTime will replace it with the following
     *
     * `%Y` - years
     * `%D` - days
     * `%H` - hours
     * `%M` - minutes
     * `%S` - seconds
     * `%m` - milliseconds
     * `%n` where n is a number,  format the following token to be padded with n zeros if all places can't be filled. E.g. a timer with 5 seconds left, the string `"%2%S"` will render as `"05"`
     */
    Timer.prototype.fmtTime = function (fmt) {
        var _this = this;
        if (fmt === void 0) { fmt = "%Y:%D:%H:%M:%S.%m"; }
        // %Y - year
        // %D - Day
        // %H - hour
        // %M - minute
        // %S - second
        // %m - millisecond
        // %n - where n is number, next format token to be padded with n zeros.
        // %% - literal percent
        return fmt.replace(/%(\d)(%\w)/g, function (p1, p2, p3) { return _this._addPadding(_this.fmtTime(p3), +p2); }).replace(/%Y/g, this.getYearsUI(0))
            .replace(/%D/g, this.getDaysUI(0))
            .replace(/%H/g, this.getHoursUI(0))
            .replace(/%M/g, this.getMinutesUI(0))
            .replace(/%S/g, this.getSecondsUI(0))
            .replace(/%m/g, this.getMillisecondsUI(0))
            .replace(/%%/g, '%');
    };
    Timer.prototype.getTime = function () {
        // TODO replace usage with toString
        return this.fmtTime();
    };
    Timer.prototype.clearTimer = function () {
        this.pause();
    };
    return Timer;
}(time_1.Time));
exports.Timer = Timer;
//# sourceMappingURL=timer.js.map