"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.clearGoodInterval = exports.setGoodInterval = exports.addPadding = void 0;
function addPadding(number, zeros) {
    var value = String(number);
    // @ts-ignore
    return "0".repeat(Math.max(zeros - value.length, 0)) + value;
}
exports.addPadding = addPadding;
var goodIntervalIds = [];
var nextId = 1;
/** A self-correcting version of [[setTimeout]], which will "speed up"
 *  subsequent calls if it falls behind (as a result of being behind too many
 *  other queued events on the event loop).
 *
 *  Can potentially cause a "burst" of the same calls, if for some reason
 *  it falls very behind.
 */
function setGoodInterval(callback, timeout) {
    var wrapper = function (callback, timeout, _lastCall, _id) {
        if (!goodIntervalIds[_id]) {
            return;
        }
        callback();
        var callTime = Date.now();
        var adjustment = callTime - _lastCall - timeout;
        var adjustedTimeout = timeout - adjustment;
        var adjustedCallTime = callTime - adjustment;
        setTimeout(function () { return wrapper(callback, timeout, adjustedCallTime, _id); }, adjustedTimeout);
    };
    var start = Date.now();
    var id = nextId++;
    goodIntervalIds[id] = true;
    setTimeout(function () { wrapper(callback, timeout, start, id); }, timeout);
    return id;
}
exports.setGoodInterval = setGoodInterval;
function clearGoodInterval(id) {
    goodIntervalIds[id] = false;
}
exports.clearGoodInterval = clearGoodInterval;
//# sourceMappingURL=index.js.map